from typing import Iterable
from graphql import ExecutionContext
import strawberry
from strawberry.extensions import SchemaExtension
from strawberry.schema.config import StrawberryConfig
from strawberry.types.scalar import ScalarDefinition, ScalarWrapper


class Schema(strawberry.federation.Schema):
    """Custom schema class to use the custom type and field functions."""

    def __init__(
        self,
        query: type | None = None,
        mutation: type | None = None,
        subscription: type | None = None,
        directives: Iterable[type] = (),
        types: Iterable[type] = (),
        extensions: Iterable[type[SchemaExtension] | SchemaExtension] = (),
        execution_context_class: type[ExecutionContext] | None = None,
        config: StrawberryConfig | None = None,
        scalar_overrides: dict[object, type | ScalarWrapper | ScalarDefinition] | None = None,
        schema_directives: Iterable[object] = (),
        enable_federation_2: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            query=query,
            mutation=mutation,
            subscription=subscription,
            directives=directives,
            types=types,
            extensions=extensions,
            execution_context_class=execution_context_class,
            config=config,
            scalar_overrides=scalar_overrides,
            schema_directives=schema_directives,
            **kwargs,
        )



from typing import Any
from django.urls import path, re_path, URLPattern
from django.conf import settings


def dynamicpath(rel_path: str, view: Any, kwargs: dict[str, Any]| None  = None, name: str | None = None) -> URLPattern:
    """ Create a path that respects the MY_SCRIPT_NAME setting. 
    
    This is useful for normal paths that need to be prefixed with the script name.
    
    Args:
        regex_path (str): The regex path to use.
        view (Any): The view to use.
        kwargs (dict[str, Any] | None, optional): Additional keyword arguments for the path.
        name (str | None, optional): The name of the path.
    Returns:
        URLPattern: The URL pattern with the script name prefixed.
    """
    if settings.MY_SCRIPT_NAME:
        return path(settings.MY_SCRIPT_NAME.lstrip("/") + "/" + rel_path,  view, kwargs, name) # type: ignore
    else:
        return path(rel_path, view, kwargs, name) # type: ignore



def re_dynamicpath(regex_path: str, view: Any, kwargs: dict[str, Any] | None = None, name: str | None = None) -> URLPattern:
    """ Create a path that respects the MY_SCRIPT_NAME setting. 
    
    This is useful for regex paths that need to be prefixed with the script name.
    
    Args:
        regex_path (str): The regex path to use.
        view (Any): The view to use.
        kwargs (dict[str, Any] | None, optional): Additional keyword arguments for the path.
        name (str | None, optional): The name of the path.
    Returns:
        URLPattern: The URL pattern with the script name prefixed.
    """
    if settings.MY_SCRIPT_NAME:
        return re_path(rf"^{settings.MY_SCRIPT_NAME.lstrip('/')}/" + regex_path.lstrip("^"), view, kwargs, name) # type: ignore
    else:
        return re_path(regex_path, view, kwargs, name) # type: ignore




from .type import (
    type,
    field,
    interface,
    django_type,
    django_field,
    django_interface,
    django_mutation,  # type: ignore[assignment]
    pydantic_input,
    mutation,
    pydantic_type,
    subscription,
    input,
    django_input,
    scalar,  # type: ignore[assignment]
    filter_type,
    filter,
    filter_field,
)
from .type import Info
from .schema import Schema

""" All public API of kante. """
__all__ = [
    "type",
    "field",
    "interface",
    "django_mutation",
    "mutation",
    "django_type",
    "django_field",
    "django_interface",
    "pydantic_type",
    "subscription",
    "input",
    "django_input",
    "scalar",
    "filter_type",
    "filter",
    "filter_field",
    "pydantic_input",
    "Info",
    "Schema",
]



from typing import (
    Callable,
    Literal,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

import strawberry
import strawberry_django
from django.db.models import Model
from strawberry.experimental import pydantic
from strawberry.federation.schema_directives import (
    Key,
)
from strawberry.types import Info
from strawberry_django.fields.field import StrawberryDjangoField
from strawberry_django.utils.typing import (
    AnnotateType,
    PrefetchType,
    TypeOrMapping,
    TypeOrSequence,
)
from strawberry_django import filters
from strawberry_django import filter_field as sfilter_field
from strawberry_django import input as sdjango_input


filter_type = filters.filter_type
filter = filters.filter
filter_field = sfilter_field

django_mutation = strawberry_django.mutation
mutation = strawberry.mutation

django_input = sdjango_input
input = strawberry.input

scalar = strawberry.scalar  # type: ignore[assignment]

interface = strawberry.interface
subscription = strawberry.subscription
type = strawberry.type
field = strawberry.field
pydantic_type = pydantic.type
pydantic_input = pydantic.input
django_field = strawberry_django.field

T = TypeVar("T", bound=object)

DjangoTypeDecorator = Callable[
    [Type[T]],
    Type[T],
]


def django_type(
    model: Type[Model],
    name: Optional[str] = None,
    field_cls: Type[StrawberryDjangoField] = StrawberryDjangoField,
    is_input: bool = False,
    is_interface: bool = False,
    is_filter: Union[Literal["lookups"], bool] = False,
    description: Optional[str] = None,
    directives: Optional[Sequence[object]] = (),
    extend: bool = False,
    filters: Optional[Type[object]] = None,
    order: Optional[Type[object]] = None,
    ordering: Optional[Type[object]] = None,
    pagination: bool = False,
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
    annotate: Optional[TypeOrMapping[AnnotateType]] = None,
    disable_optimization: bool = False,
    fields: Optional[Union[list[str], Literal["__all__"]]] = None,
    exclude: Optional[list[str]] = None,
    federated: bool = True,
) -> Callable[
    [Type[T]],
    Type[T],
]:
    if federated:
        directives = list(directives or [])
        directives.append(Key(fields="id"))

    def wrapper(cls: Type[T]) -> Type[T]:
        """A decorator to create a Django type with federation support."""

        if federated:
            # Check if id field is defined in type annotations
            annotations = getattr(cls, "__annotations__", {})
            assert "id" in annotations, (
                "Django type must have an 'id' field annotation for federation"
            )

            # Check if resolve_reference method is defined in the class
            # Note: kante federation will add this if not present
            if not hasattr(cls, "resolve_reference"):
                # Add a default resolve_reference method that looks up by id
                def resolve_reference(cls: Type[object], info: Info, id: str) -> object:
                    return model.objects.aget(id=id)

                setattr(cls, "resolve_reference", classmethod(resolve_reference))

        return strawberry_django.type(
            model,
            name=name,
            field_cls=field_cls,
            is_input=is_input,
            is_interface=is_interface,
            is_filter=is_filter,
            description=description,
            directives=directives,
            extend=extend,
            filters=filters,
            order=order,
            ordering=ordering,
            pagination=pagination,
            only=only,
            select_related=select_related,
            prefetch_related=prefetch_related,
            annotate=annotate,
            disable_optimization=disable_optimization,
            fields=fields,
            exclude=exclude,
        )(cls)

    return wrapper


def django_interface(
    model: Type[Model],
    name: Optional[str] = None,
    field_cls: Type[StrawberryDjangoField] = StrawberryDjangoField,
    description: Optional[str] = None,
    directives: Optional[Sequence[object]] = (),
) -> Callable[[Type[T]], Type[T]]:
    """Decorator to create a Django interface type."""

    def wrapper(cls: Type[T]) -> Type[T]:
        """A decorator to create a Django interface type."""
        return strawberry_django.interface(
            model,
            name=name,
            field_cls=field_cls,
            description=description,
            directives=directives,
        )(cls)

    return wrapper
